from .computation_engine import (apply, join, summarize, fullstream,
                                 apply_local, join_local, summarize_local, fullstream_local,
                                 apply_debug, join_debug, summarize_debug, fullstream_debug,
                                 source,
                                 Engine, output, DelayedException,
                                 Source,
                                 WorkflowError
                                 )

from .fs_cache import (CacheBuilder, LocalCacheBuilder, LazyCacheBuilder, CacheItem, AnnotatedItem,
                       InvalidCacheException,  sigterm_protected_open, sigterm_protection,
                       sigterm_protected,
                       CacheListItem)

from .std_logger import get_logger

# DO NOT TOUCH THE FOLLOWING LINE:
import sys
if not getattr(sys, "frozen", False):
    import pkg_resources  # part of setuptools
    __version__ = tuple(map(int, pkg_resources.require(__name__)[0].version.split(".")))
else:
    __version__ = (0, 0, 0)



# encoding: utf-8

import contextlib
import functools
import glob
import hashlib
import inspect
import json
import multiprocessing
import os
import signal
import sys
import time
import threading
import weakref

import dill
import jsonpickle


from .std_logger import get_logger


class AtomicCounter(object):

    lock = threading.Lock()

    def __init__(self):
        self.value = 0

    def inc(self):
        with AtomicCounter.lock:
            self.value += 1

    def dec(self):
        with AtomicCounter.lock:
            self.value -= 1

    def is_zero(self):
        return self.value == 0


class sigterm_protection(object):

    """Context manager which blocks SIGTERM until the context is left.
    A caught sigterm will be recorded and handled with sys.exit() as soon as no other
    sigterm_protection context manager is active.
    """

    opened_files = AtomicCounter()

    """
    we need the counter because we may only restore the signal handler if no other thread
    redirected this using this context manager !
    """

    def __init__(self, *a, **kw):
        self.a = a
        self.kw = kw

    def __enter__(self, *a, **kw):
        self._term_triggered = False

        def record_sigterm(sig, frame):
            get_logger().warn("caught SIGTERM during file I/O")
            get_logger().warn("args where %s and %s" % (self.a, self.kw))
            self._term_triggered = True

        self._orig_sigterm_handler = signal.signal(signal.SIGTERM, record_sigterm)
        self.opened_files.inc()

    def __exit__(self, *a, **kw):
        signal.signal(signal.SIGTERM, self._orig_sigterm_handler)
        self.opened_files.dec()
        if self._term_triggered:
            while not self.opened_files.is_zero():
                time.sleep(0.001)
            get_logger().warn("exit because of caught SIGTERM during file I/O")
            sys.exit(128 + signal.SIGTERM)  # convention


@contextlib.contextmanager
def sigterm_protected_open(*a, **kw):
    """SIGTERM protected open() function. For more details read the doc string of
    sigterm_protection()
    """
    with sigterm_protection(*a, **kw):
        yield open(*a, **kw)


def sigterm_protected(fun):
    """function decorator which creates a SIGTERM protected function. For more details read the doc
    string of sigterm_protection()
    """
    @functools.wraps(fun)
    def inner(*a, **kw):
        with sigterm_protection(*a, **kw):
            return fun(*a, **kw)
    return inner


def json_pretty_encode(obj):
    json_str = jsonpickle.encode(obj)
    pretty_json_str = json.dumps(json.loads(json_str), indent=4)
    return pretty_json_str


def _fallback_save_function():
    ext = ".pickled"

    def save_function(what, path):
        with sigterm_protected_open(path, "wb") as fp:
            dill.dump(what, fp)
    return ext, save_function


def _fallback_load_function():

    def load_function(path):
        with sigterm_protected_open(path, "rb") as fp:
            return dill.load(fp)
    return load_function


def rel_to_cache_root(p):
    """keep parentf folder + file given for file path p, eg /abc/cde/xdf/x.table -> xdf/x.table"""
    return os.path.join(os.path.basename(os.path.split(p)[0]), os.path.split(p)[1])


class CacheLoadError(Exception):

    pass


class CacheItemLoadError(Exception):

    pass


class InvalidCacheException(Exception):

    pass


class CacheItem(object):

    __slots__ = ["annotation", "hash_code_args", "hash_code_result", "rel_path", "loaded", "value",
                 "last_path", "injected"]

    def __init__(self, hash_code_args, hash_code_result, annotation, injected):
        self.hash_code_args = hash_code_args
        self.hash_code_result = hash_code_result
        self.annotation = annotation.copy()
        self.injected = injected
        self.rel_path = None
        self.loaded = False
        self.value = None
        self.last_path = None

    def __eq__(self, other):
        for name in self.__slots__:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    def __str__(self):
        return "<CachedItem annotation=%s rel_path=%s value=%r loaded=%s hash_code_args=%s hash_code_result=%s>" % (
            self.annotation, self.rel_path, self.value, self.loaded, self.hash_code_args, self.hash_code_result)

    def print_(self, fh=sys.stdout):
        print >> fh, json_pretty_encode(self)

    def load_result(self, cb):
        __, f_ext = os.path.splitext(self.rel_path)
        load_function = cb.tr.lookup_load_function_for(f_ext)
        used_customized_load_function = load_function is not None
        if load_function is None:
            load_function = _fallback_load_function()
        path = os.path.join(cb.root_folder, self.rel_path)
        try:
            with sigterm_protection():
                obj = load_function(path)
        except Exception:
            used = "used %s load function" % ("customized" if used_customized_load_function else "fallback")
            raise CacheLoadError("could not load cache content from %s, %s" % (path, used))
        return obj

    def dump(self, path):
        self.last_path = rel_to_cache_root(path)
        with sigterm_protected_open(path, "wb") as fh:
            fh.write(json_pretty_encode(self))

    def _check_attributes(self):
        missing = [name for name in self.__slots__ if not hasattr(self, name)]
        if missing:
            msg = ", ".join(missing)
            raise InvalidCacheException("attributes %s are missing. Better clean the cache" % msg)

    @staticmethod
    def load(path):
        try:
            with sigterm_protected_open(path, "rb") as fh:
                data = fh.read()
        except Exception:
            raise IOError("reading from %s failed" % path)
        try:
            item = jsonpickle.decode(data)
        except ValueError:
            raise CacheItemLoadError("data in %s is not valid JSON format, maybe your cache is "
                                     "invalid to a pacer update, deleting the cache might help" % fh.name)
        item._check_attributes()
        item.last_path = rel_to_cache_root(fh.name)
        return item


class AnnotatedItem(object):

    def __init__(self, value, **annotation):
        self.value = value
        self.annotation = annotation


class CacheListItem(CacheItem):

    __slots__ = CacheItem.__slots__ + ["sub_items"]

    def __init__(self, *a):
        super(CacheListItem, self).__init__(*a)
        self.sub_items = []

    def __str__(self):
        return "<CacheListItem annotation=%s rel_path=%s value=%r loaded=%s hash_code_args=%s hash_code_result=%s>" % (
            self.annotation, self.rel_path, self.value, self.loaded, self.hash_code_args, self.hash_code_result)

    def create_save_function(self, cb):
        """
        We want to keep CacheItems very minimal for pickling, which is the reason that they have
        no _CacheBuilder attributes.
        Instead we create here the save function depending on cb on the fly.
        """

        def save(what, path):
            self.sub_items = []
            for i, item in enumerate(what):
                ext, save_function = cb.tr.lookup_ext_and_save_function_for(item)
                if ext is None or save_function is None:
                    ext, save_function = _fallback_save_function()
                full_path = path + "_%d" % i + ext
                with sigterm_protection():
                    save_function(item, full_path)
                hash_code = cb.tr.compute_hash(item)
                sub_item = CacheItem(self.hash_code_args, hash_code, self.annotation, self.injected)
                sub_item.rel_path = rel_to_cache_root(full_path)
                self.sub_items.append(sub_item)
            self.rel_path = rel_to_cache_root(path)

        return save

    def load_result(self, cb):
        items = []
        for p in glob.glob(os.path.join(cb.root_folder, self.rel_path + "_*.*")):
            __, f_ext = os.path.splitext(p)
            load_function = cb.tr.lookup_load_function_for(f_ext)
            if load_function is None:
                load_function = _fallback_load_function()
            with sigterm_protection():
                obj = load_function(p)
            items.append(obj)
        return items

    def __iter__(self):
        return iter(self.sub_items)


class DistributedTypeRegistry(object):

    def __init__(self):
        self._pickler = dill.dumps
        self._unpickler = dill.loads
        self.reset_handlers()

    def reset_handlers(self):
        self._type_handlers = list()

    def register_handler(self, type_, hash_data_source, file_extension, load_function,
                         save_function, info_getter):
        self._type_handlers.append((type_,
                                    self._pickler(hash_data_source),
                                    file_extension,
                                    self._pickler(load_function),
                                    self._pickler(save_function),
                                    self._pickler(info_getter),
                                    )
                                   )

    def lookup_load_function_for(self, ext):
        for (__, __, ext_i, load_function, __, __) in self._type_handlers:
            if ext == ext_i:
                return self._unpickler(load_function)
        return None

    def _lookup_for_type_of(self, obj):
        for row in self._type_handlers:
            if isinstance(obj, row[0]):
                return row
        nonf = self._pickler(None)
        return [None, nonf, None, nonf, nonf, nonf, nonf]

    def lookup_hash_data_extractor_for(self, key):
        return self._unpickler(self._lookup_for_type_of(key)[1])

    def lookup_ext_and_save_function_for(self, what):
        row = self._lookup_for_type_of(what)

        ext = row[2]
        save_function = self._unpickler(row[4])

        return ext, save_function

    def lookup_info_getter(self, what):
        row = self._lookup_for_type_of(what)
        return self._unpickler(row[5])

    def compute_hash(self, key):
        return self._compute_hash(key, set(), [])

    def _compute_hash(self, key, seen, keep):
        """we use see to handle recursive data strucutres, which might happen for example
        for namedtuples sometimes...
        further we keep local data in a list so that the id() calls are unique (we had
        different tuples here with same id because they were created temporarily and
        when recursing they were destroyed and other tuples occupied the same memory
        location, thus resulted in the same id)
        """
        if isinstance(key, CacheItem):
            return key.hash_code_result
        keep.append(key)
        extractor = self.lookup_hash_data_extractor_for(key)
        if extractor is not None:
            key = extractor(key)
            keep.append(key)
        if id(key) in seen and not isinstance(key, (int, float, bool, long, str, tuple)):
            # stop recursion here, return a placeholder which is unique for the item:
            # len() is a good choide because it correspondst to the current state of the
            # stack:
            data = "___memo_%s" % len(seen)
        else:
            seen.add(id(key))
            if isinstance(key, str):
                data = key
            elif isinstance(key, unicode):
                data = key.encode("utf-8")
            elif isinstance(key, (bool, int, long, float,)):
                data = str(dill.dumps(key))
            elif key is None:
                data = "__None__"
            elif isinstance(key, (tuple, list)):
                data = []
                for item in key:
                    keep.append(item)
                    data.append(self._compute_hash(item, seen, keep))
                data = "".join(data)
            elif isinstance(key, set):
                data = []
                for item in sorted(key):
                    keep.append(item)
                    data.append(self._compute_hash(item, seen, keep))
                data = "".join(data)
            elif isinstance(key, dict):
                data = []
                for item in key.items():
                    keep.append(item[0])
                    keep.append(item[1])
                    keep.append(item)
                    data.append(self._compute_hash(item, seen, keep))
                data = "".join(data)
            elif hasattr(key, "__dict__"):
                data = self._compute_hash(key.__dict__, seen, keep)
            else:
                raise Exception("can not compute hash for %r" % key)
            if not isinstance(data, basestring):
                raise RuntimeError("implementation error: data should be str, but is %s" % type(data))
        muncher = hashlib.sha1()
        muncher.update(data)
        hash_ = muncher.hexdigest()
        return hash_


class LocalTypeRegistry(DistributedTypeRegistry):

    def __init__(self):
        noop = lambda x: x
        self._pickler = noop
        self._unpickler = noop
        self.reset_handlers()


class _CacheBuilder(object):

    def __init__(self, root_folder=None):
        self.root_folder = root_folder
        self._logger = get_logger(self)
        self._functions = []

    def register_handler(self, *a, **kw):
        self.tr.register_handler(*a, **kw)

    def _setup_folder(self, function):
        folder = function.__name__
        if self.root_folder is not None:
            folder = os.path.join(self.root_folder, folder)
        return folder

    def __call__(self, function):
        # folder = self._setup_folder(function)
        cache, lock, counter = self._setup_cache_internals()

        clz = self._cache_function_class()
        c = clz(function, self.root_folder, lock, cache, counter, self.tr)

        self._functions.append(c)
        return c

    def inject_result(self, item, value):
        for f in self._functions:
            new_item = f.inject_result(item, value)
            if new_item is not None:  # hit !
                return new_item

    def remove_injected_result(self, item):
        for f in self._functions:
            f.remove_injected_result(item)

    def _cache_function_class(self):
        return _CachedFunction

    def create_dict(self):
        return dict()

    def remove_unused_entries(self):
        to_keep = set()
        for path in glob.glob(os.path.join(self.root_folder, "*", "__*")):
            to_keep.add(os.path.basename(path)[2:])
        for path in glob.glob(os.path.join(self.root_folder, "*", "*.cache_entry")):
            # we look for proper cache entries, we might have other files in the cache
            # folder...
            hash_value = os.path.splitext(os.path.basename(path))[0]
            if hash_value not in to_keep:
                try:
                    os.remove(path)
                except IOError, e:
                    self._logger.error("could not remove %s: %s" % (path, e.message))

        used_refs = set()
        for path in glob.glob(os.path.join(self.root_folder, "*", "*.cache_entry")):
            item = CacheItem.load(path)
            referenced_path = os.path.join(self.root_folder, item.rel_path)
            used_refs.add(referenced_path)
            if isinstance(item, CacheListItem):
                for sub_item in item:
                    referenced_path = os.path.join(self.root_folder, sub_item.rel_path)
                    used_refs.add(referenced_path)

        for path in glob.glob(os.path.join(self.root_folder, "*", "*")):
            basename = os.path.basename(path)
            if not basename.startswith("__") and not path.endswith(".cache_entry"):
                if path not in used_refs:
                    try:
                        os.remove(path)
                    except IOError, e:
                        self._logger.error("could not remove %s: %s" % (path, e.message))

    def remove_use_markers(self):
        for path in glob.glob(os.path.join(self.root_folder, "*", "__*")):
            try:
                os.remove(path)
            except IOError, e:
                self._logger.error("could not remove %s: %s" % (path, e.message))


class CacheBuilder(_CacheBuilder):

    def __init__(self, root_folder=None):
        super(CacheBuilder, self).__init__(root_folder)
        self.tr = DistributedTypeRegistry()
        self._manager = multiprocessing.Manager()

        # shutdown manager if object is deleted, this has less impact to garbage collector
        # than implementing __del__:
        def on_die(ref, manager=self._manager, logger=get_logger()):
            logger.info("try to shutdown multiprocessings manager process")
            manager.shutdown()
            logger.info("finished shutdown multiprocessings manager process")
        self._del_ref = weakref.ref(self, on_die)

    def _setup_cache_internals(self):
        cache = self._manager.dict()
        lock = self._manager.Lock()
        counter = self._manager.Value('d', 0)
        return cache, lock, counter

    def __str__(self):
        return "<CacheBuilder(%s)>" % self.root_folder

    def create_dict(self):
        return self._manager.dict()


class LazyCacheBuilder(CacheBuilder):

    def _cache_function_class(self):
        return _LazyCachedFunction

    def __str__(self):
        return "<LazyCacheBuilder(%s)>" % self.root_folder


class LocalCounter(object):

    def __init__(self):
        self.value = 0


class NoOpContextManager(object):

    def __enter__(self, *a, **kw):
        pass

    __exit__ = __enter__


class LocalCacheBuilder(_CacheBuilder):

    """Cache which only resists in current process, can not be used with pacerd distributed
    computation capabilities ! Use CacheBuilder instead.
    """

    def __init__(self, root_folder=None):
        super(LocalCacheBuilder, self).__init__(root_folder)
        self.tr = LocalTypeRegistry()
        self._manager = None

    def _setup_cache_internals(self):
        cache = dict()
        lock = NoOpContextManager()
        counter = LocalCounter()
        return cache, lock, counter  # , handlers

    def __str__(self):
        return "<LocalCacheBuilder(%s)>" % self.root_folder


class _CachedFunction(object):

    """ Instances of this class can be used to decorate function calls for caching their
    results, even if the functions are executed across different processes started by Python
    multiprocessing modules Pool class.

    The cache is backed up on disk, so that cache entries are persisted over different
    runs.
    """

    def __init__(self, function, root_folder, _lock, _cache, _counter, tr):

        self.function = function
        self.__name__ = function.__name__

        self._cache = _cache
        self._lock = _lock
        self._hit_counter = _counter

        self.root_folder = root_folder

        self.tr = tr

        self._logger = get_logger(self)
        self._setup_cache()
        self._arg_processor = None

    def folder(self):
        return os.path.join(self.root_folder, self.function.__name__)

    def __getstate__(self):
        dd = self.__dict__.copy()
        if "_logger" in dd:
            del dd["_logger"]
        return dd

    def register_handler(self, *a, **kw):
        self.tr.register_handler(*a, **kw)

    def __setstate__(self, dd):
        self.__dict__.update(dd)
        self._logger = get_logger(self)

    def get_number_of_hits(self):
        return self._hit_counter.value

    def set_cache_key_preprocessor(self, **kw):
        self._arg_processor = kw

    def _setup_cache(self):
        folder = os.path.join(self.root_folder, self.function.__name__)
        if not os.path.exists(folder):
            os.makedirs(folder)
        for file_name in os.listdir(folder):
            path = os.path.join(folder, file_name)
            stem, ext = os.path.splitext(file_name)
            if ext == ".cache_entry":
                self._cache[stem] = CacheItem.load(path)

    def clear(self):
        self._cache.clear()

    def _contains(self, hash_code):
        return hash_code in self._cache.keys()

    def _get(self, hash_code):
        item = self._cache[hash_code]
        if not item.loaded:
            value = item.load_result(self)
            item.value = value
            item.loaded = True
        self._cache[hash_code] = item
        return item.value

    def _put(self, hash_code_args, result, annotation, injected):
        assert isinstance(annotation, dict)
        item = self._store(result, hash_code_args, annotation, injected)
        self._cache[hash_code_args] = item
        self._set_use_marker(hash_code_args)
        return item

    def inject_result(self, item, value):
        ci = self._cache.get(item.hash_code_args)
        if ci is not None:
            new_item = self._put(ci.hash_code_args, value, item.annotation, True)
            return new_item

    def remove_injected_result(self, item):
        ci = self._cache.get(item.hash_code_args)
        if ci is not None:
            os.remove(os.path.join(self.root_folder, ci.last_path))
            del self._cache[ci.hash_code_args]

    def _store(self, result, hash_code_args, annotation, injected):

        hash_code_result = self.tr.compute_hash(result)

        # create cache entry
        if isinstance(result, list):
            item = CacheListItem(hash_code_args, hash_code_result, annotation, injected)
            save_function = item.create_save_function(self)
            ext = ""
        else:
            item = CacheItem(hash_code_args, hash_code_result, annotation, injected)
            ext, save_function = self.tr.lookup_ext_and_save_function_for(result)
            if ext is None or save_function is None:
                ext, save_function = _fallback_save_function()

        # write result if not known
        path = os.path.join(self.root_folder, self.function.__name__, hash_code_result + ext)
        if not os.path.exists(path):
            with sigterm_protection():
                save_function(result, path)
            self._logger.info("stored %s" % path)
        else:
            self._logger.info("no need to store item to %s" % path)

        item.rel_path = rel_to_cache_root(path)
        self._write_cache_item(hash_code_args, item)
        return item

    def _write_cache_item(self, hash_code_args, item):
        name = hash_code_args + ".cache_entry"
        path = os.path.join(self.root_folder, self.function.__name__, name)
        item.dump(path)
        return item

    def _get_names(self, args):
        for arg in args:
            if isinstance(arg, CacheItem):
                yield arg.name
            else:
                getter = self.tr.lookup_info_getter(arg)
                if getter is not None:
                    yield getter(arg)

    def _setup_args(self, args):
        if self._arg_processor is None:
            return args
        arg_names = inspect.getargspec(self.function).args
        result = []
        for arg, name in zip(args, arg_names):
            if not isinstance(arg, CacheItem):
                if name in self._arg_processor:
                    arg_processor = self._arg_processor.get(name)
                    if hasattr(arg_processor, "__call__"):
                        arg = arg_processor(arg)
                    else:
                        arg = arg_processor
            result.append(arg)
        return result

    def cached_call(self, args, kw):
        self._logger.info("cached call for %s" % self.function.__name__)
        all_args = list(args) + list(sorted(kw.items()))
        try:
            args_for_hash = self._setup_args(all_args)
            hash_code_args = self.tr.compute_hash(args_for_hash)
        except RuntimeError:
            raise Exception("could not compute hash for %r. maybe you should register your own "
                            "handler" % (all_args,))
        if self._contains(hash_code_args):
            with self._lock:
                self._hit_counter.value += 1
            self._logger.info("cache hit for %s" % hash_code_args)
            ci = self._get(hash_code_args)
            # ci.print_()
            self._set_use_marker(hash_code_args)
            return ci, ci

        args = self.resolve_inputs(args)
        result = self.function(*args, **kw)


        annotation = {}
        if isinstance(result, AnnotatedItem):
            annotation = result.annotation
            result = result.value

        self._logger.info("store new result for %s" % hash_code_args)
        item = self._put(hash_code_args, result, annotation, False)
        self._logger.info("stored new result for %s" % hash_code_args)
        return result, item

    def _set_use_marker(self, hash_code):
        sub_folder = os.path.join(self.root_folder, self.function.__name__)
        marker_path = os.path.join(sub_folder, "__" + hash_code)
        try:
            with sigterm_protected_open(marker_path, "w"):
                pass
        except IOError, e:
            self._logger.error("could not write to %s: %s" % (marker_path, e.message))
            raise

    def __call__(self, *args, **kw):
        result, item = self.cached_call(args, kw)
        return result

    def resolve_inputs(self, i):
        return i

    def __str__(self):
        return "<_CachedFunction(%s)>" % self.function.__name__


class _LazyCachedFunction(_CachedFunction):

    def _get(self, hash_code):
        item = self._cache[hash_code]
        # item ist multiple + modues is "beak lists" ?
        # aufliösen zu liste von cacheitems
        return item

    def __call__(self, *args, **kw):
        result, item = self.cached_call(args, kw)
        return item

    def __str__(self):
        return "<_LazyCachedFunction(%s)>" % self.function.__name__

    def resolve_inputs(self, args):
        if isinstance(args, (list, tuple)):
            loaded = [a.load_result(self) if isinstance(a, CacheItem) else self.resolve_inputs(a) for a in args]
            if isinstance(args, (tuple,)):
                loaded = tuple(loaded)
            return loaded
        if isinstance(args, CacheItem):
            return args.load_result(self)
        return args



try: import readline  # For readline input support
except: pass

import sys, os, traceback, signal, codeop, cStringIO, cPickle, tempfile

def pipename(pid):
    """Return name of pipe to use"""
    return os.path.join(tempfile.gettempdir(), 'debug-%d' % pid)

class NamedPipe(object):
    def __init__(self, name, end=0, mode=0666):
        """Open a pair of pipes, name.in and name.out for communication
        with another process.  One process should pass 1 for end, and the
        other 0.  Data is marshalled with pickle."""
        self.in_name, self.out_name = name +'.in',  name +'.out',
        try: os.mkfifo(self.in_name,mode)
        except OSError: pass
        try: os.mkfifo(self.out_name,mode)
        except OSError: pass
        
        # NOTE: The order the ends are opened in is important - both ends
        # of pipe 1 must be opened before the second pipe can be opened.
        if end:
            self.inp = open(self.out_name,'r')
            self.out = open(self.in_name,'w')
        else:
            self.out = open(self.out_name,'w')
            self.inp = open(self.in_name,'r')
        self._open = True

    def is_open(self):
        return not (self.inp.closed or self.out.closed)
        
    def put(self,msg):
        if self.is_open():
            data = cPickle.dumps(msg,1)
            self.out.write("%d\n" % len(data))
            self.out.write(data)
            self.out.flush()
        else:
            raise Exception("Pipe closed")
        
    def get(self):
        txt=self.inp.readline()
        if not txt: 
            self.inp.close()
        else:
            l = int(txt)
            data=self.inp.read(l)
            if len(data) < l: self.inp.close()
            return cPickle.loads(data)  # Convert back to python object.
            
    def close(self):
        self.inp.close()
        self.out.close()
        try: os.remove(self.in_name)
        except OSError: pass
        try: os.remove(self.out_name)
        except OSError: pass

    def __del__(self):
        self.close()
        
def remote_debug(sig,frame):
    """Handler to allow process to be remotely debugged."""
    def _raiseEx(ex):
        """Raise specified exception in the remote process"""
        _raiseEx.ex = ex
    _raiseEx.ex = None
    
    try:
        # Provide some useful functions.
        locs = {'_raiseEx' : _raiseEx}
        locs.update(frame.f_locals)  # Unless shadowed.
        globs = frame.f_globals
        
        pid = os.getpid()  # Use pipe name based on pid
        pipe = NamedPipe(pipename(pid))
    
        old_stdout, old_stderr = sys.stdout, sys.stderr
        txt = ''
        pipe.put("Interrupting process at following point:\n" + 
               ''.join(traceback.format_stack(frame)) + ">>> ")
        
        try:
            while pipe.is_open() and _raiseEx.ex is None:
                line = pipe.get()
                if line is None: continue # EOF
                txt += line
                try:
                    code = codeop.compile_command(txt)
                    if code:
                        sys.stdout = cStringIO.StringIO()
                        sys.stderr = sys.stdout
                        exec code in globs,locs
                        txt = ''
                        pipe.put(sys.stdout.getvalue() + '>>> ')
                    else:
                        pipe.put('... ')
                except:
                    txt='' # May be syntax err.
                    sys.stdout = cStringIO.StringIO()
                    sys.stderr = sys.stdout
                    traceback.print_exc()
                    pipe.put(sys.stdout.getvalue() + '>>> ')
        finally:
            sys.stdout = old_stdout # Restore redirected output.
            sys.stderr = old_stderr
            pipe.close()

    except Exception:  # Don't allow debug exceptions to propogate to real program.
        traceback.print_exc()
        
    if _raiseEx.ex is not None: raise _raiseEx.ex
    
def debug_process(pid):
    """Interrupt a running process and debug it."""
    os.kill(pid, signal.SIGUSR1)  # Signal process.
    pipe = NamedPipe(pipename(pid), 1)
    try:
        while pipe.is_open():
            txt=raw_input(pipe.get()) + '\n'
            pipe.put(txt)
    except EOFError:
        pass # Exit.
    pipe.close()

def listen():
    signal.signal(signal.SIGUSR1, remote_debug) # Register for remote debugging.

if __name__=='__main__':
    if len(sys.argv) != 2:
        print "Error: Must provide process id to debug"
    else:
        pid = int(sys.argv[1])
        debug_process(pid)



# encoding: utf-8


import tblib.pickling_support
# allows pickling of basic information from tracebacks:
tblib.pickling_support.install()

import Queue
import abc
import dill
import functools
import itertools
import multiprocessing
import multiprocessing.pool
import os
import random
import sys
import thread
import time
import traceback

from .std_logger import get_logger

from .fs_cache import CacheListItem


class TerminationAwarePool(multiprocessing.pool.Pool):

    def __init__(self, *a, **kw):
        super(TerminationAwarePool, self).__init__(*a, **kw)
        self.terminated = False

    def terminate(self):
        self.terminated = True
        super(TerminationAwarePool, self).terminate()



class DelayedException(object):

    """store exception for throwing it later.
       we use this for transfering an exception on a worker process to the
       main process.
       only works because we imported and installed tblib above
       """

    def __init__(self, ee):
        self.ee = ee
        __,  __, self.tb = sys.exc_info()

    def re_raise(self):
        raise self.ee, None, self.tb

    def as_string(self):
        return "".join(traceback.format_exception(self.ee, None, self.tb, limit=None))


class WorkflowInterruption(DelayedException):

    def as_string(self):
        return str(self.ee)


class WorkflowError(Exception):

    pass


def delayed_exception(e):

    if isinstance(e, WorkflowError):
        return WorkflowInterruption(e)
    return DelayedException(e)


def printable(what):
    what = repr(what)
    if len(what) > 60:
        return what[:28] + " .. " + what[-28:]
    else:
        return what


def eval_pickled(f_s, args_s, raise_exceptions=False):
    """ requires function and args serialized with dill, so we can remote execute even lambdas or
    decorated functions or static methods, which 'pure' multiprocessing can not handle"""

    try:
        args = dill.loads(args_s)
    except Exception, e:
        get_logger(eval_pickled).error("got exception when unpickling args: {}".format(e))
        if raise_exceptions:
            raise e
        return delayed_exception(e)
    try:
        f = dill.loads(f_s)
    except Exception, e:
        get_logger(eval_pickled).error("got exception when unpickling f: {}".format(e))
        if raise_exceptions:
            raise e
        return delayed_exception(e)
    try:
        s_args = printable(args)
        get_logger(f).debug("start {}{} in process {}".format(f.__name__, s_args, os.getpid()))
        result = f(*args)
        s_result = printable(result)
        get_logger(f).debug("got result {!s} from process {}".format(s_result, os.getpid()))
        return result
    except Exception, e:
        get_logger(f).error("got exception: {}".format(e))
        get_logger(f).exception(e)
        if raise_exceptions:
            raise e
        return delayed_exception(e)


class EnumeratedItem(object):

    def __init__(self, number, value):
        self.number = number
        self.value = value

    def __iter__(self):
        return iter((self.number, self.value))


def wrap_callback(new_enumeration, callback):
    """creates a new callback which builds a EnumeratedItem from the reult """
    def new_callback(result):
        callback(EnumeratedItem(new_enumeration, result))
    return new_callback


class Engine(object):

    def __init__(self, n_workers, raise_exceptions=False):

        self.pool = None
        self.n = n_workers
        self.raise_exceptions = raise_exceptions

    def _setup_pool(self):

        n = self.n
        if n is not None:
            if n < 0:
                n = multiprocessing.cpu_count() - n
                n = max(n, 0)
            if n > 0:
                self.pool = TerminationAwarePool(n)
            elif n == 0:
                self.pool = None
            else:
                raise Exception("can not set number of processes to %d" % n)

    def _run_async(self, f, args, callback, run_local=False, debug=False):
        """ f is not implemented for handling enumerated items, so we exctract the items
        and construct a 'new enumeration' for the result.
        This is needed in order to ensure that the final result of the computations have
        a fixed order although their computation order might be different due to the
        concurrent processing.

        in debug mode the computation runs local but goes through the pickling/unpickling loop
        this enables use of debugger and allows detection of pickling problems

        """
        # we use eval_pickled in both cases to make shure we have the same kind of error handling
        # for local and remote execution. this supports debugging.

        if debug:
            # run computation local so we can use debugger
            # but use the pickling / unpickling procedure to detect pickling problems
            f_s = dill.dumps(f)
            args_s = dill.dumps(args)
            result = eval_pickled(f_s, args_s, self.raise_exceptions)
            callback(result)
        else:
            run_local = run_local is True or self.pool is None
            if run_local:
                try:
                    get_logger(self).debug("start {} with args {}".format(f, printable(args)))
                    result = f(*args)
                    get_logger(self).debug("function {} returned {}".format(f, printable(result)))
                except Exception, e:
                    get_logger(self).debug("function {} rose {}".format(f, e))
                    if os.environ.get("DEBUG", False):
                        import pdb
                        import sys
                        import traceback
                        type, value, tb = sys.exc_info()
                        traceback.print_exc()
                        pdb.post_mortem(tb)
                    if self.raise_exceptions:
                        raise e
                    result = delayed_exception(e)
                callback(result)
            else:
                f_s = dill.dumps(f)
                args_s = dill.dumps(args)
                get_logger(self).debug("send to pool %s %s %s" % (f, args, callback))
                if not self.pool.terminated:
                    self.pool.apply_async(eval_pickled, (f_s, args_s), callback=callback)
                else:
                    get_logger(self).debug("no apply_async")
                    callback(None)

    def _start(self, node):
        self._setup_pool()
        self.exception = None
        self.pool_created = True
        node._start_computations(self)

    def shut_down(self):
        self.pool.close()
        self.pool.join()

    @staticmethod
    def run_once(node, n_workers, re_raise_exceptions=False, continuation_callback=None):
        e = Engine(n_workers)
        try:
            result = e.run(node, re_raise_exceptions, continuation_callback)
            return result
        finally:
            e.shut_down()

    def run(self, node, re_raise_exceptions=False, continuation_callback=None):
        l = get_logger(self)
        l.info("")
        l.info("*" * 80)
        l.info("START NEW COMPUTATION")
        l.info("*" * 80)
        l.info("")
        if continuation_callback is not None:
            assert self.n not in (0, None), "you must use worker procecces for this"
            def check():
                while True:
                    if not continuation_callback():
                        self.terminate()
                        node.stop()
                        print("stopped")
                        return
                    time.sleep(.01)
            thread.start_new(check, ())
        self._start(node)
        return self._fetch_results(node, re_raise_exceptions)

    def _fetch_results(self, node, re_raise_exceptions):
        try:
            results = node._get_all_in_order(re_raise_exceptions)
            return results
        finally:
            if self.pool is not None:
                self.pool.close()
                self.pool.join()

    def terminate(self):
        if self.n in (0, None):
            raise Exception("kill() only works for workflows running on multiple "
                            "processes")
        self._terminate()

    def _terminate(self):
        # if we kill to fast, pool is not setup, so we wait:
        while self.pool is None:
            time.sleep(0.001)
        get_logger(self).info("ask to kill process pool")
        self.pool.terminate()
        self.pool.join()
        get_logger(self).info("process pool killed")


def _info(listener):
    return listener.f if hasattr(listener, "f") else listener


def _stop_item():
    msg = "workflow stopped due to user request"
    exception = DelayedException(Exception(msg))
    return EnumeratedItem(-1, exception)


class DeadlyPill(object):

    def __init__(self, payload):
        self.payload = payload


def is_deadly_pill(item):
    return isinstance(item, DeadlyPill)


class DataStream(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._queues = dict()
        self.stopped = False

    def stop(self):
        self.stopped = True
        get_logger(self).info("%s workflow execution termination requested", self)

    def empty(self, listener):
        get_logger(self).debug("%s asks for empty input queue", _info(listener))
        return self._queues.get(listener).empty()

    def get(self, listener):
        while True:
            try:
                item = self._queues.get(listener).get(timeout=.01)
                break
            except Queue.Empty:
                if self.stopped:
                    return _stop_item()
            if self.engine.exception is not None:
                return self.engine.exception

        get_logger(self).debug("%s asks and gets item %d with value %s", _info(listener),
                               item.number, item.value)
        return item

    @abc.abstractmethod
    def size(self):
        pass

    def put(self, item):
        """this is where computation results are passed to."""
        get_logger(self).debug("put %s" % item)
        for listener, queue in self._queues.items():
            get_logger(self).debug("put item %d with value %s to %s",
                                   item.number, item.value, _info(listener))
            queue.put(item)

    def register_listener(self, listener):
        assert listener not in self._queues
        self._queues[listener] = Queue.Queue()

    def _start_computations(self, engine):
        """
        starts computations at leafs first then up to the root
        """
        self.engine = engine
        get_logger(self).debug("%s._start_computations() called ", self)
        for stream in self.input_streams:
            stream._start_computations(engine)
        self._start_computation()
        return self


class Source(DataStream):

    def __init__(self, items=None):
        super(Source, self).__init__()
        self.items = items
        self.input_streams = []

    def set_input_items(self, items):
        self.items = items

    def set_input_item(self, item):
        self.items = [item]

    def size(self):
        return len(self.items)

    def _start_computation(self):
        get_logger(self).debug("%s._start_computation() called ", self)
        for i, item in enumerate(self.items):
            if self.stopped:
                self.put(EnumeratedItem(i, WorkflowInterruptioWorkflowInterruption("user stopped computation")))
                return
            self.put(EnumeratedItem(i, item))


class Computation(DataStream):

    def __init__(self, f, inputs, run_local=False, debug=False):
        super(Computation, self).__init__()
        self.run_local = run_local
        self.debug = debug
        self.f = f
        self.input_streams = self._setup_inputs(inputs)

    def stop(self):
        super(Computation, self).stop()
        for stream in self.input_streams:
            stream.stop()

    def _setup_inputs(self, inputs):
        streams = []
        for input_ in inputs:
            stream = input_
            if not isinstance(input_, DataStream):
                if isinstance(input_, (tuple, list)):
                    stream = Source(input_)
                else:
                    stream = Source((input_,))
            stream.register_listener(self)
            streams.append(stream)
        return streams

    @abc.abstractmethod
    def _start_computation(self):
        pass

    def _start_async_computation_for(self, result_number, args):
        assert all(isinstance(arg, EnumeratedItem) for arg in args)
        forward_result = wrap_callback(result_number, self.put)
        get_logger(self).debug("_start_async_computation_for")
        for arg in args:
            if isinstance(arg.value, DelayedException):
                self.engine.exception = arg.value
                forward_result(arg.value)
                return

        if self.stopped:
            forward_result(WorkflowInterruption("workflow stopped by user"))
            return

        args = tuple(a for (__, a) in args)
        get_logger(self).debug("Engine._run_async %s %s" % (self.f, args))
        self.engine._run_async(self.f, args, forward_result, self.run_local, self.debug)


class OutputNode(DataStream):

    def __init__(self, input_):
        super(OutputNode, self).__init__()

        if not isinstance(input_, DataStream):
            if isinstance(input_, (tuple, list)):
                input_ = Source(input_)
            else:
                input_ = Source((input_,))

        input_.register_listener(self)
        self.input_ = input_

    def size(self):
        return self.input_.size()

    def __str__(self):
        return "<OutputNode(id=%d, %s)>" % (id(self), self.input_)

    def stop(self):
        super(OutputNode, self).stop()
        self.input_.stop()

    def _start_computations(self, engine, do_not_reraise_exceptions=True):
        """
        starts computations at leafs first then up to the root
        """
        self.engine = engine
        self.input_._start_computations(engine)
        get_logger(self).debug("%s.start_computation() called ", self)

        self.results = None
        results = []
        while len(results) < self.size():
            item = self.input_.get(listener=self)
            if is_deadly_pill(item):
                return
            if not do_not_reraise_exceptions:
                if isinstance(item.value, DelayedException):
                    if not isinstance(item.value, WorkflowInterruption):
                        item.value.re_raise()
            #if self.engine.exception is not None:
                #if not do_not_reraise_exceptions:
                    #raise self.engine.exception
                #else:
                    #results.append(DelayedException(self.engine.exception))
                    #continue

            results.append(item)

        results.sort(key=lambda en_item: en_item.number)
        self.results = [item.value for item in results]

    def _get_all_in_order(self, re_raise_exceptions=False):
        while self.results is None:
            time.sleep(0.01)
        if re_raise_exceptions:
            for r in self.results:
                if isinstance(r, DelayedException):
                    if not isinstance(r, WorkflowInterruption):
                        r.re_raise()
        return self.results

    def still_running(self):
        return self.results is None


class _UpdateManagerForJoin(object):

    """ manages input data from sources which might come in in arbitrary order.  handles
    incremental generation of input_streams for computations.  """

    def __init__(self, input_streams):
        self.seen = set()
        self.data = [[None for __ in range(a.size())] for a in input_streams]
        self.next_ = [0] * len(input_streams)

    def new_function_arguments_for(self, stream_index, item):
        self.data[stream_index][self.next_[stream_index]] = item
        self.next_[stream_index] += 1
        # cross product over all indices of available data:
        sizes = [len(v) for v in self.data]
        for perm in itertools.product(*(range(size) for size in sizes)):
            if perm not in self.seen:
                args = tuple(self.data[j][pi] for j, pi in enumerate(perm))
                if all(ai is not None for ai in args):
                    number = args[0].number
                    for ai, si in zip(args[1:], sizes[1:]):
                        number = si * number + ai.number
                    yield number, args
                    self.seen.add(perm)

    def input_data_pending(self, size):
        return len(self.seen) < size


class JoinComputation(Computation):

    def size(self):
        return reduce(lambda x, y: x * y, (a.size() for a in self.input_streams))

    def __str__(self):
        return "<JoinComputation(id=%d, %s)>" % (id(self), self.f)

    def _start_computation(self):
        get_logger(self).debug("%s._start_computation() called ", self)

        manager = _UpdateManagerForJoin(self.input_streams)
        while manager.input_data_pending(self.size()):
            started_computation = False
            for i, stream in enumerate(self.input_streams):
                if not stream.empty(listener=self):
                    item = stream.get(listener=self)
                    if is_deadly_pill(item):
                        return
                    if isinstance(item.value, DelayedException):
                        self.put(item)
                        return
                    for number, args in manager.new_function_arguments_for(i, item):
                        self._start_async_computation_for(number, args)
                        started_computation = True
            if not started_computation:
                time.sleep(0.001)
        get_logger(self).debug("%s._start_computation() finished ", self)


class FetchAllComputation(Computation):

    def _collect_input_data(self):
        items = []
        stream = self.input_streams[0]
        for i in range(stream.size()):
            item = stream.get(listener=self)
            if is_deadly_pill(item):
                return None, None
            if isinstance(item.value, DelayedException):
                for _ in range(self.size()):
                    self.put(item)
                return None, None
            items.append(item)
            #if self.engine.exception is not None:
                #for _ in range(self.size()):
                    #self.put(DelayedException(self.engine.exception))
                #return None, None

        fixed_args = []
        for other_stream in self.input_streams[1:]:
            assert other_stream.size() == 1
            item = other_stream.get(listener=self)
            if is_deadly_pill(item):
                return
            if isinstance(item.value, DelayedException):
                for _ in range(self.size()):
                    self.put(item)
                return None, None
            fixed_args.append(item)

        items.sort(key=lambda item: item.number)
        values = [v for __, v in items]
        return values, fixed_args


class SummarizeComputation(FetchAllComputation):

    def size(self):
        return 1

    def __str__(self):
        return "<SummarizeComputation(id=%d, %s)>" % (id(self), self.f)

    def _start_computation(self):
        get_logger(self).debug("%s._start_computation() called ", self)

        values, fixed_args = self._collect_input_data()
        if values is None:
            return
        arg = EnumeratedItem(0, values)
        args = (arg,) + tuple(fixed_args)
        self._start_async_computation_for(0, args)


class FullStreamComputation(FetchAllComputation):

    def size(self):
        return self.input_streams[0].size()

    def __str__(self):
        return "<FullStreamComputation(id=%d, %s)>" % (id(self), self.f)

    def _start_computation(self):
        get_logger(self).debug("%s._start_computation() called ", self)

        values, fixed_args = self._collect_input_data()
        if values is None:
            return

        # this uses the plain engine because we hava a special callback which
        # creates a stream from a single computed list
        # this is why we have to work with plain data, not EnumeratedItem instances:
        def forward_result(result):
            get_logger(self).debug("forward %s" % result)
            if isinstance(result, (list, CacheListItem)):
                for i, item in enumerate(result):
                    self.put(EnumeratedItem(i, item))
            elif isinstance(result, DelayedException):
                for i in range(self.size()):
                    self.put(EnumeratedItem(i, result))

            else:
                c = result.__class__
                raise ValueError("result of a fullstream computation must be a list or "
                                 "CacheListItem, got %s !" % c)

        args = (values,) + tuple(a.value for a in fixed_args)
        self.engine._run_async(self.f, args, forward_result, self.run_local, self.debug)


class ComputedSource(Computation):

    def __init__(self, f, args):
        super(ComputedSource, self).__init__(f, [], True, False)
        self.f = f
        self.args = args
        self.items = None

    def size(self):
        return len(self.items)

    def __str__(self):
        return "<ComputedSource(id=%d, %s)>" % (id(self), self.f)

    def _start_computation(self):
        get_logger(self).debug("%s._start_computation() called ", self)
        args = self.args
        if len(args) == 1 and isinstance(args[0], Source):
            args = args[0].items
            assert len(args) == 1, "exepected source which only produces one data item"
        else:
            args = self.args
        self.items = list(self.f(*args))
        for i, item in enumerate(self.items):
            self.put(EnumeratedItem(i, item))


class ZipComputation(Computation):

    """either input_streams provide all n inputs or one fixed input. The latter is managed by
    ConstantSource class.
    """

    def size(self):
        sizes = set(s.size() for s in self.input_streams)
        if 1 in sizes:
            sizes.remove(1)
        if len(sizes) > 1:
            msg = "allow ownly sources with same sized outputs or constant constant source"
            raise Exception(msg)
        if len(sizes) == 1:
            return sizes.pop()
        # sizes has size 0 means: we only had inputs of size 1
        return 1

    def __str__(self):
        return "<ZipComputation(id=%d, %s)>" % (id(self), self.f)

    def _start_computation(self):
        get_logger(self).debug("%s._start_computation() called ", self)
        # wait and get the constant args which we need anyway:
        cip = self._collect_data_from_constant_input_streams()
        if cip is None:
            return

        if self._found_and_handled_exception_in(cip):
            return

        # only constant inputs ?
        if self.size() == 1:
            self._start_async_computation_for(0, cip)
        else:
            self._zip_all(cip)

    def _collect_data_from_constant_input_streams(self):
        constant_inputs = [None] * len(self.input_streams)
        for i, stream in enumerate(self.input_streams):
            if stream.size() == 1:
                item = stream.get(listener=self)
                if is_deadly_pill(item):
                    return
                constant_inputs[i] = item
        return constant_inputs

    def _found_and_handled_exception_in(self, args):
        for item in args:
            if item is not None:
                if isinstance(item.value, DelayedException):
                    for _ in range(self.size()):
                        self.put(item)
                    return True
        return False

    def _zip_all(self, cip):
        seen = set()
        n = 0
        input_matrix = self._setup_input_matrix(cip)
        while n < self.size()  and not self.stopped:
            avail = self._check_for_any_new_data()
            if avail:
                for number, args in self._assemble_function_arguments(input_matrix, seen):
                    if self._found_and_handled_exception_in(args):
                        return
                    self._start_async_computation_for(number, args)
                    n += 1
            else:
                time.sleep(0.001)

    def _setup_input_matrix(self, cip):
        input_matrix = [[None for __ in self.input_streams] for k in range(self.size())]
        for i, is_ in enumerate(self.input_streams):
            if is_.size() == 1:
                for ii in range(self.size()):
                    input_matrix[ii][i] = cip[i]
        return input_matrix

    def _check_for_any_new_data(self):
        for stream in self.input_streams:
            if stream.size() > 1:
                if not stream.empty(listener=self):
                    return True
        return False

    def _assemble_function_arguments(self, input_matrix, seen):
        for i, stream in enumerate(self.input_streams):
            if not stream.empty(listener=self):
                item = stream.get(listener=self)
                if is_deadly_pill(item):
                    return
                input_matrix[item.number][i] = item

        for ni, arg in enumerate(input_matrix):
            if ni not in seen:
                if all(c is not None for c in arg):
                    yield ni, arg
                    seen.add(ni)


def create_decorator(clz, debug, run_local):
    def decorator(inner):
        @functools.wraps(inner)
        def wrapped(*args):
            return clz(inner, args, debug=debug, run_local=run_local)
        wrapped.inner = inner
        return wrapped
    return decorator


apply = create_decorator(ZipComputation, False, False)
apply_local = create_decorator(ZipComputation, False, True)
apply_debug = create_decorator(ZipComputation, True, False)

join = create_decorator(JoinComputation, False, False)
join_local = create_decorator(JoinComputation, False, True)
join_debug = create_decorator(JoinComputation, True, False)

summarize = create_decorator(SummarizeComputation, False, False)
summarize_local = create_decorator(SummarizeComputation, False, True)
summarize_debug = create_decorator(SummarizeComputation, True, False)

fullstream = create_decorator(FullStreamComputation, False, False)
fullstream_local = create_decorator(FullStreamComputation, False, True)
fullstream_debug = create_decorator(FullStreamComputation, True, False)


def source(inner):
    @functools.wraps(inner)
    def wrapped(*args):
        return ComputedSource(inner, args)
    wrapped.inner = inner
    return wrapped


def output(inner):
    if isinstance(inner, (Computation)):
        return OutputNode(inner)

    @functools.wraps(inner)
    def wrapped(*args):
        return OutputNode(inner(*args))
    wrapped.inner = inner
    return wrapped


if __name__ == "__main__":

    @join
    def add(*args):
        time.sleep(0.5 * random.random())
        return sum(args)

    @apply
    def inc(x, config=dict()):
        time.sleep(0.5 * random.random())
        return x + config.get("increment", 1)

    @summarize
    def avg(values):
        return float(sum(values)) / len(values)

    Engine.set_number_of_processes(7)
    N = 3

    s1 = range(N)
    s2 = range(1, N + 1)
    r = add(0, 1, s1)

    r = inc(avg(inc(add(0, 1, add(inc(s1), inc(s2), inc(7))), dict(increment=2))))

    r = output(r)
    r._start_computations()

    print(sorted(r.get_all_in_order()))
